/*
 *	File:	CalcBrot.cp
 *	---------------------
 *
 *	This file actually implements the Mandelbrot calculations.
 *	At the root of this file is the first function you'll see
 *	which calculates a Mandelbrot point.  The important thing
 *	to keep in mind is that the Thread Manager is used extensively
 *	to help keep the user interactiveness alive.
 */

#include "ThreadBrot.h"

OSPixelPtr	ThreadBrot::CalcBrotPoint( OSPixelPtr offscreenptr, int h, int v, Complex *c)
{
register _Float			z, zi, t, ti, z2, z2i;
register _Float			cc = c->real, cci = c->imag;
	
	long			loop;

	z2 = z = cc;
	z2i = zi = cci;

	for( loop = 0; loop < iterations && 
			((z2 *= z) + (z2i *= zi)) <= 4; loop++)
	{
		ti = z * zi;
		t = z2 - z2i;
		ti += ti;
		z2 = z = t + cc;
		z2i = zi = ti + cci;
		
		if( !pointCount--)
			if( !(offscreenptr = Yield( h, v)))
				return nil;
	}
	
	PlotPoint( offscreenptr, loop);
	
	return offscreenptr;
}

void	ThreadBrot::PlotPoint( OSPixelPtr p, long v)
{
	if( (p < osStart) || (p > osEnd))
		DebugStr( "\pOUT OF BOUNDS ERROR");	
	else
		if( sizeof(OSPixel) == sizeof( long))
		{
			if( v == iterations)
				v = 255;
			
			else
				v &= 255;
		
			v = 255 - v;
		
			*p = (OSPixel) (v | v << 8 | v << 16);
		}
		else if( sizeof(OSPixel) == sizeof( char))
		{
			if( v == iterations)
				v = 255;
			
			else
				v %= 250;
		
			*p = (OSPixel) v;
		}
}



void	ThreadBrot::FrameBrot( void)
{
	Complex			cpt = topLeftC;
	Point			p = topLeftP;
	short			width = botRightP.h-topLeftP.h;
	short			height = botRightP.v-topLeftP.v;
	_Float			left;
	
	DrawRow( &cpt, &p, width);
	DrawColumn( &cpt, &p, height);

	left = cpt.imag;
	cpt.imag -= scale * height;
	p.v += height -1;
	DrawRow( &cpt, &p, width);

	cpt.imag = left;
	cpt.real += width * scale;
	p = topLeftP;
	p.h += width -1;
	DrawColumn( &cpt, &p, height);
}


void	ThreadBrot::CalcRect( void)
{
	short			width = botRightP.h-topLeftP.h;
	short			height = botRightP.v-topLeftP.v;
	short			j, h, v;
	OSPixelPtr		offscreen;
	Complex			cpt = topLeftC;
	long			th, tv;
	
	if( !width || !height)
		return;
	
	h = topLeftP.h +1;
	v = topLeftP.v +1;

	offscreen = GeneratePixels( h, v);

	UpdateRect( &botRightP);

	cpt.imag -= scale;
	cpt.real += scale;

	while( --width)
	{
		for( j = height -1; j; j--)
		{
			offscreen = CalcBrotPoint( offscreen, h, v, &cpt);
			offscreen += osRowPixels;
			v++;
			cpt.imag -= scale;
		}
		offscreen -= osRowPixels * (height -1);
		++offscreen;
		cpt.imag = topLeftC.imag - scale;
		cpt.real += scale;
		v = topLeftP.v +1;
		h++;
	}
}


void	ThreadBrot::FillRect( void)
{
	short			width = botRightP.h-topLeftP.h;
	short			height = botRightP.v-topLeftP.v;
	short			j, h, v;
	OSPixelPtr		offscreen, tempOSptr;
	OSPixel			pix;
	
	if( !width || !height)
		return;
	
	h = topLeftP.h +1;
	v = topLeftP.v +1;

	tempOSptr = offscreen = GeneratePixels( h, v);
	pix = *(offscreen - osRowPixels -1);

	UpdateRect( &botRightP);

	while( --width)
	{
		for( j = height -1; j; j--)
		{
			*offscreen = pix;
			offscreen += osRowPixels;
		}
		offscreen = ++tempOSptr;
	}
}


Boolean	ThreadBrot::TraceEdges( void)
{
	OSPixelPtr			topRow, btmRow, leftCol, rightCol;
	OSPixel				testByte;
	short				length;

	topRow = GeneratePixels( &topLeftP);
	btmRow = GeneratePixels( topLeftP.h, botRightP.v);

	testByte = *topRow;
	
	for( length = botRightP.h - topLeftP.h; length; length--)
		if( *topRow++ != testByte || *btmRow++ != testByte)
			return false;
	
	leftCol = GeneratePixels( &topLeftP);
	rightCol = GeneratePixels( botRightP.h, topLeftP.v);
	
	for( length = botRightP.v - topLeftP.v; length; length--)
	{
		if( *leftCol != testByte || *rightCol != testByte)
			return false;
		leftCol += osRowPixels;
		rightCol += osRowPixels;
	}
	
	return true;
}


void ThreadBrot::DrawColumn( Complex *topC, Point *topP, short length)
{
	Complex			z = *topC;
	short			i;
	OSPixelPtr		offscreenptr;
	long			tv, th;
	long			stackSize;
	
	stackSize = gTopOfStack - (long) &z;
	
	if( stackSize > gMaxLengthOfStack)
		gMaxLengthOfStack = stackSize;
	
	
	offscreenptr = GeneratePixels( topP);
	
	for( i = 0; i < length; i++)
	{
		tv = dv;	th = dh;
		offscreenptr = CalcBrotPoint( offscreenptr, topP->h, topP->v+i, &z);
		if( th != dh || tv != dv)
			if( topP->h + dh >= osImageArea.right ||
				topP->h + dh < osImageArea.left)
					break;
			else if( topP->v + dv + length -1 >= osImageArea.bottom ||
					 topP->v + dv < osImageArea.top)
					 {
					 	DebugStr( "\pThis should never happen");
					 	if( dv > tv)
							length -= dv - tv;
					}
		
		offscreenptr += osRowPixels;
		z.imag -= scale;
	}
	UpdateRect( topP->h, topP->v+length);
}


void ThreadBrot::DrawRow( Complex *leftC, Point *leftP, short length)
{
	Complex			z = *leftC;
	short			i;
	OSPixelPtr		offscreenptr;
	long			th, tv;
		
	offscreenptr = GeneratePixels( leftP);
	
	for( i = 0; i < length; i++)
	{
		th = dh; tv = dv;
		offscreenptr = CalcBrotPoint( offscreenptr, leftP->h+i, leftP->v, &z);
		if( th != dh || tv != dv)
			if( leftP->h + dh + i >= osImageArea.right ||
				leftP->h + dh + i < osImageArea.left)
					if( dh > th)
						length -= dh - th;
			else if( leftP->v + dv >= osImageArea.bottom ||
					 leftP->v + dv < osImageArea.top)
						break;

		++offscreenptr;
		z.real += scale;
	}
	UpdateRect( leftP->h+length, leftP->v);
}


