/*
 *	File:	ThreadBrot.cp
 *	---------------------
 *
 *	This file implements the structures which support the generation
 *	of cool mandelbrot images.  CalcBrot.cp actually does all the work.
 *	The important thing to keep in mind is that the Thread Manager
 *	is used extensively to help keep the user interactiveness alive.
 */

#include "ThreadBrot.h"

/*
 *				( int xx, int yy, register int ww, int hh,
 *				register _Float rr, register _Float ii, register _Float ss, 
 *				GWorldPtr os, WindowPtr wind, register int iter)
 *			: iterations(iter), osImage(os), scale(ss),dx(0),dy(0),
 *			  yieldTime( 2), window(wind)
 */
	ThreadBrot::ThreadBrot( ParametericsPtr p)
{
	OSErr			errWhatErr;
	ThreadID		tempThreadInfo;
	short			tryTwice;
	
	osImage 		= p->osImage;
	osImageArea 	= p->osImageArea;
	window 			= p->window;
	queue			= p->queue;
	topLeftC 		= p->topLeftC;
	topLeftP 		= p->topLeftP;
	botRightP		= p->botRightP;
	scale 			= p->scale;
	bitDepth 		= p->bitDepth;
	iterations 		= p->iterations;
	frame 			= p->frame;

	dh = 0;
	dv = 0;
	
	pointCount = newPointCount = kNewPointCount;
	
	bigPicture		= window->portRect;

	threadRect.left	= topLeftP.h;
	threadRect.right = botRightP.h;
	threadRect.top = topLeftP.v;
	threadRect.bottom = botRightP.v;
	threadC			= topLeftC;
	
	queue->QueueThread( this);
	
	queue->TitleIsNumThrds( p->window);
	
	tryTwice = 2;
	while( tryTwice--)
	{
		errWhatErr = NewThread( kCooperativeThread, 
								ThreadBrotEntry, 
								(void *)this, 
								50000,
								kCreateIfNeeded,
								(void**)nil,
								&tempThreadInfo);
		if( errWhatErr == noErr)
			break;
		
	}		
	if( errWhatErr)
		DebugStr( "\pXcuzzU?  NewThread Failed??");
	
	threadInfo = tempThreadInfo;
	
	YieldToThread( threadInfo);

}


void	ThreadBrot::OpenOSImage( void)
{
	osPixmap = GetGWorldPixMap( osImage);
	
	LockPixels( osPixmap);
	
	pixels = (OSPixelPtr)GetPixBaseAddr( osPixmap);
	
	osRowBytes = (**osPixmap).rowBytes & 0x3FFF;
	osRowPixels = osRowBytes / sizeof(OSPixel);

	osStart = pixels;
	osEnd = &pixels[osImageArea.bottom * osRowPixels - 1];

	startTime = TickCount();

	updateRect.top		= osImageArea.bottom;
	updateRect.left		= osImageArea.right;
	updateRect.bottom	= osImageArea.top;
	updateRect.right	= osImageArea.left;
}


void	ThreadBrot::MarkRect( Rect *newRect)
{
	return;
	
	GrafPtr			savePort;
	
	GetPort( &savePort);
	SetPort( window);
	PaintRect( newRect);
	SetPort( savePort);
}


void	ThreadBrot::OffsetThread( int offsetH, int offsetV)
{
	ThreadBrot		*replacementThread;
	Parameterics	p;
	Rect			newImageArea;
	Boolean			overLap;
	
/*
	newImageArea.top = topLeftP.v;
	newImageArea.left = topLeftP.h;
	newImageArea.right = botRightP.h;
	newImageArea.bottom = botRightP.v;
 */
 	newImageArea = threadRect;

	OffsetRect( &newImageArea, offsetH + dh, offsetV + dv);
	overLap = SectRect( &newImageArea, &bigPicture, &newImageArea);

	MarkRect( &newImageArea);
	
	if( overLap)
	{
		dh += offsetH;
		dv += offsetV;
		
		if( (threadRect.right	- threadRect.left ==
			newImageArea.right  - newImageArea.left) &&
			(threadRect.bottom	- threadRect.top ==
			newImageArea.bottom - newImageArea.top))
				return;

		SetupParameterics( &p);

		p.topLeftC			= threadC;
		
		if( offsetH < 0 && 
			threadRect.right	- threadRect.left ==
			newImageArea.right  - newImageArea.left)
				p.topLeftC.real -= scale * offsetH;
		
		if( offsetV < 0 &&
			threadRect.bottom	- threadRect.top ==
			newImageArea.bottom - newImageArea.top)
				p.topLeftC.imag += scale * offsetV;
		
		p.topLeftP.h		= newImageArea.left;
		p.topLeftP.v		= newImageArea.top;
		p.botRightP.h		= newImageArea.right;
		p.botRightP.v		= newImageArea.bottom;
		p.frame 			= true;
		
		replacementThread = new ThreadBrot( &p);
	}

	queue->DeQueueThread( this);
	queue->TitleIsNumThrds( window);
	DisposeThread( threadInfo, nil, true);
}

void	ThreadBrot::CloseOSImage( void)
{
	GrafPtr		savedPort;
	Rect		tempUpdateRect = updateRect;
	
	osPixmap = GetGWorldPixMap( osImage);

	GetPort( &savedPort);
	SetPort( window);

	OffsetRect( &tempUpdateRect, dh, dv);
	
	CopyBits( (const BitMap *) (*(osImage->portPixMap)), 
			  (const BitMap *) (*(((CWindowPtr)window)->portPixMap)),
			  &tempUpdateRect,&tempUpdateRect,srcCopy,nil);
//			  &bigPicture,&bigPicture,srcCopy,nil);
			  
	
	
	SetPort( savedPort);

	UnlockPixels( osPixmap);
}


void	ThreadBrot::CalculateNewPointCount( void)
{
	long			deltaTime = TickCount() - startTime;
	
	if( deltaTime)
		newPointCount = (int) (((_Float) newPointCount) * 
//					((_Float) kYieldTime / (_Float) deltaTime));
					((_Float) 30 / (_Float) deltaTime));
	else
		newPointCount = newPointCount * 6 + 1;
	
	pointCount = newPointCount;
}


pascal void	*ThreadBrotEntry( void *there)
{
	ThreadBrot		*t = (ThreadBrot *)there;

	t->Entry();
	
	return nil;
}


long ggLargestStack = 0;

void	ThreadBrot::Entry( void)
{
	long		aTopLocatedStack;
	
	gTopOfStack = (long) &aTopLocatedStack;
	gMaxLengthOfStack = 0;
	
	OpenOSImage();
	if( frame)
/*
	{
		while( topLeftP.h < botRightP.h)
		{
 */
			FrameBrot();
/*
			topLeftC.imag -= scale;
			topLeftC.real += scale;
			topLeftP.h++;
			topLeftP.v++;
			botRightP.h--;
			botRightP.v--;
		}
	}
 */
	TestRect();
	CloseOSImage();
	queue->DeQueueThread( this);
	queue->TitleIsNumThrds( window);
	ggLargestStack = gMaxLengthOfStack;
}


OSPixelPtr	ThreadBrot::Yield( int h, int v)
{
	CalculateNewPointCount();
	CloseOSImage();
	YieldToAnyThread();
	OpenOSImage();

	return GeneratePixels( h, v);
}

OSPixelPtr	ThreadBrot::GeneratePixels( Point *p)
{
	UpdateRect( p->h, p->v);	

	if( p->h + dh >= osImageArea.right ||
		p->h + dh < osImageArea.left)
			return nil;
	else if( p->v + dv >= osImageArea.bottom ||
			 p->v + dv < osImageArea.top)
			return nil;

	return &pixels[ (p->v + dv) * osRowPixels + p->h + dh];
}


OSPixelPtr	ThreadBrot::GeneratePixels( int h, int v)
{
	UpdateRect( h, v);

	if( h + dh >= osImageArea.right ||
		h + dh < osImageArea.left)
			return nil;
	else if( v + dv >= osImageArea.bottom ||
			 v + dv < osImageArea.top)
			return nil;

	return &pixels[ (v + dv) * osRowPixels + h + dh];
}

void ThreadBrot::UpdateRect( Point *p)
{
	UpdateRect( p->h, p->v);
}

void ThreadBrot::UpdateRect( int h, int v)
{
	if( h < updateRect.left)
		updateRect.left = h;
	if( updateRect.right < h)
		updateRect.right = h +1;
	if( v < updateRect.top)
		updateRect.top = v;
	if( updateRect.bottom < v)
		updateRect.bottom = v +1;
}





